"""High-level functions."""


import json
from typing import Union, TextIO

from smart_open import open

from .compile import compile
from .context import ResolutionContext


_empty_context = ResolutionContext()


def run(file: Union[str, TextIO], context: ResolutionContext = _empty_context):
    """Run a named task specification file and return the resulting object."""
    if isinstance(file, str):
        with open(file, "r") as f:
            raw_task = json.load(f)
    else:
        raw_task = json.load(file)
    task_def = compile(raw_task)
    return task_def.resolve(context)



"""Functions to extract Definitions from JSON-like objects representation."""


from functools import singledispatch
import collections.abc
from typing import Any, Mapping

from .defs import (Definition, ConcreteValue, NameReference, Call, SeqDef,
                   DictDef, ContextValue, ContextBinder)
from .magics import get_magic


def _compile_dict_vals(mapping: Mapping[Any, Any]) -> Mapping[Any, Definition]:
    return {key: compile(val) for key, val in mapping.items()}


@singledispatch
def compile(obj: Any) -> Definition:
    """Extract a definition from a JSON-like object representation."""
    return ConcreteValue(obj)

@compile.register(list)
@compile.register(tuple)
def _(obj):
    return SeqDef(list, map(compile, obj))

@compile.register(collections.abc.Mapping)
def _(obj):
    if len(obj) == 1:
        key, val = next(iter(obj.items()))
        if key.startswith("!"):
            key = key[1:]
            if key:
                arg = compile(val)
            magic = get_magic(key)
            if magic:
                pass_context = getattr(magic, "_pass_context", False)
                return Call(ConcreteValue(magic), [arg],
                            pass_context=1 if pass_context else None)
            elif key:
                return Call(NameReference(key), [arg])
        elif key == "=":
            return ContextValue(val)
    elif "=" in obj:
        obj = obj.copy()
        values = _compile_dict_vals(obj.pop("="))
        wrapped = compile(obj)
        return ContextBinder(values, wrapped)
    elif "!" in obj:
        obj = obj.copy()
        callee = obj.pop("!")
        posargs = list(map(compile, obj.pop("*", [])))
        kwargs = _compile_dict_vals(obj)
        return Call(NameReference(callee), posargs, kwargs)
    else:
        return DictDef(_compile_dict_vals(obj))



"""Basic object definitions."""


from abc import ABC, abstractmethod
from pydoc import locate
from typing import Any, Iterable, Sequence, Mapping, Type, Union

from .context import ResolutionContext


class Definition(ABC):
    """Base class for definitions of Python objects to be created."""

    @abstractmethod
    def resolve(self, context: ResolutionContext):
        pass

    def __call__(self, context: ResolutionContext):
        return self.resolve(context)


class ConcreteValue(Definition):
    """Definition wrapping an actual Python object."""

    __slots__ = "obj"

    def __init__(self, obj: Any):
        self.obj = obj

    def resolve(self, context: ResolutionContext):
        return self.obj


class NameReference(Definition):
    """Reference by dotted path to a Python object."""

    __slots__ = "name"

    def __init__(self, name: str):
        self.name = name

    def resolve(self, context: ResolutionContext):
        return locate(self.name)


class Call(Definition):
    """Definition of object obtained by callable invocation."""

    __slots__ = "callee", "posargs", "kwargs", "pass_context"

    def __init__(self, callee: Definition, posargs: Iterable[Definition] = (),
                 kwargs: Mapping[str, Definition] = {},
                 pass_context: Union[None, int, str] = None):
        self.callee = callee
        self.posargs = tuple(posargs)
        self.kwargs = dict(kwargs)
        self.pass_context = pass_context

    def resolve(self, context: ResolutionContext):
        callee = self.callee.resolve(context)
        posargs = [arg.resolve(context) for arg in self.posargs]
        kwargs = {name: arg.resolve(context)
                  for name, arg in self.kwargs.items()}
        if isinstance(self.pass_context, int):
            posargs.insert(self.pass_context, context)
        elif isinstance(self.pass_context, str):
            kwargs[self.pass_context] = context
        return callee(*posargs, **kwargs)


class SeqDef(Definition):
    """Definition of a sequence with values expressed as definitions."""

    __slots__ = "type", "items"

    def __init__(self, type: Type[Sequence], items: Iterable[Definition]):
        self.type = type
        self.items = tuple(items)

    def resolve(self, context: ResolutionContext):
        return self.type(item.resolve(context) for item in self.items)


class DictDef(Definition):
    """Definition of a dictionary with values expressed as definitions."""

    __slots__ = "items"

    def __init__(self, items: Mapping[Any, Definition]):
        self.items = items

    def resolve(self, context: ResolutionContext):
        return {key: val.resolve(context) for key, val in self.items.items()}


class ContextValue(Definition):
    """Reference to a context-bound value."""

    __slots__ = "name"

    def __init__(self, name: str):
        self.name = name

    def resolve(self, context: ResolutionContext):
        return context.vals[self.name]


class ContextBinder(Definition):
    """Wraps a definition with added context-bound values."""

    __slots__ = "values", "wrapped"

    def __init__(self, values: Mapping[str, Definition], wrapped: Definition):
        self.values = dict(values)
        self.wrapped = wrapped

    def resolve(self, context: ResolutionContext):
        resolved_vals = {key: val.resolve(context)
                         for key, val in self.values.items()}
        inner_context = context.update_vals(resolved_vals)
        return self.wrapped.resolve(inner_context)



from typing import Any, Mapping


class ResolutionContext:
    """Holds contextual information accessible during definition resolution."""

    def __init__(self, vals: Mapping[str, Any] = {}):
        self.vals = dict(vals)

    def copy(self):
        res = ResolutionContext(
            vals=self.vals.copy()
        )
        return res

    def update_vals(self, new_vals: Mapping[str, Any]) -> "ResolutionContext":
        res = self.copy()
        res.vals.update(new_vals)
        return res



"""Magic functions available in task specifications."""


import pickle

from smart_open import open


def get_magic(name):
    """Get a magic function by name or None if not exists."""
    return globals().get("do_" + name)


def accepts_context(fun):
    fun._pass_context = True
    return fun


@accepts_context
def do_run(filename, context):
    """Runs the task specified in the given file."""
    from .api import run   # here to avoid circular import
    return run(filename, context)


def do_unpickle(filename):
    """Unpickles an object from a given file."""
    with open(filename, "rb") as f:
        return pickle.load(f)



from setuptools import setup, find_packages
import os
import re

basedir = os.path.abspath(os.path.dirname(__file__))

def file(*path):
    return os.path.join(basedir, *path)

def get_long_description():
    with open(file("README.md"), "r") as f:
        return f.read()

def find_version():
    with open(file("spytex", "__init__.py"), "r") as f:
        source = f.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", source, re.M)
    if not match:
        raise RuntimeError("cannot find version string")
    return match.group(1)

setup(
    name="spytex",
    version=find_version(),
    author="rrobby86",
    author_email="rrobby86@gmail.com",
    description="Run arbitrary Python functions indicated by JSON specs",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/rrobby86/spytex",
    packages=find_packages(),
    install_requires=[
        "smart-open>=1.8.1",
    ],
    entry_points={
        "console_scripts": ["spytex=spytex.cli:main"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import pickle
from argparse import ArgumentParser
import sys

from smart_open import open

from .api import run


def main():
    parser = ArgumentParser()
    parser.add_argument("task_file", nargs="?",
                        help="task specification file")
    parser.add_argument("-p", "--pickle", metavar="FILE",
                        help="pickle returned object to specified file")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="suppress output of returned value")
    parser.add_argument("--version", action="store_true",
                        help="print version information and exit")
    args = parser.parse_args()
    if args.version:
        from . import __version__
        print("SPyTEx {}".format(__version__))
        return
    result = run(args.task_file or sys.stdin)
    if result is not None and not args.quiet:
        print(result)
    if args.pickle:
        with open(args.pickle, "wb") as f:
            pickle.dump(result, f)
